"""
Predicate for use within Forseti
"""
# pylint: disable=too-few-public-methods

class Symbol(object):
    """
    Symbols, smallest logical unit. Only accepts strings to label them
    """
    name = "Symbol"
    arity = 0

    def __init__(self, arg):
        if not isinstance(arg, str):
            raise TypeError(str(arg) + " is not a string")
        self.arg = arg
        self. args = [arg]

    def __str__(self):
        return self.arg

    def __repr__(self):
        """
        see __repr__()
        """
        return self.arg

    def __eq__(self, other):
        """
        Test if a symbol is equal with another symbol (otherwise False)
        """
        if isinstance(other, Symbol):
            return self.arg == other.arg
        else:
            return False

    def __ne__(self, other):
        """
        Test if symbol is not equal with another symbol (otherwise True)
        """
        return not self == other

    def __lt__(self, other):
        if isinstance(other, LogicalOperator):
            return True
        elif isinstance(other, Symbol):
            return self.arg < other.arg
        else:
            return NotImplemented

    def __le__(self, other):
        return self < other or self == other

    def __gt__(self, other):
        return other < self

    def __ge__(self, other):
        return other < self or other == self


class LogicalOperator(object):
    """
    Base LogicalOperator class. LogicalOperators are "Not", "Or", etc that
    take in true/false values and would return a true/false value
    """
    name = ""
    arity = 0

    def __init__(self, *kwargs):
        """

        :type kwargs: list (LogicalOperator | Symbol)
        """
        self.args = []
        kwargs = sorted(kwargs)
        for kwarg in kwargs:
            if not isinstance(kwarg, LogicalOperator) and not isinstance(kwarg, Symbol):
                raise TypeError(str(kwarg) + " is not an Operator or Symbol")

    def __str__(self):
        raise NotImplementedError("Not implemented yet")

    def __repr__(self):
        """
        Return __repr__() for predicate

        :return: __repr__()
        """
        raise NotImplementedError("Not implemented yet")

    def __eq__(self, other):
        if isinstance(self, type(other)):
            if self.arity == other.arity:
                for i in range(self.arity):
                    if self.args[i] != other.args[i]:
                        return False
                return True
        return False

    def __ne__(self, other):
        return not self == other

    def __lt__(self, other):
        if isinstance(other, Symbol):
            return False
        elif isinstance(other, LogicalOperator):
            if isinstance(self, type(other)):
                for i in range(self.arity):
                    if not self.args[i] < other.args[i]:
                        return False
                return True
        else:
            return NotImplemented

    def __le__(self, other):
        return self < other or self == other

    def __gt__(self, other):
        return other < self

    def __ge__(self, other):
        return other < self or other == self


class Not(LogicalOperator):
    """
    Not (negation) Predicate
    """
    name = "not"
    arity = 1

    def __init__(self, arg):
        super(Not, self).__init__(arg)
        self.args.append(arg)

    def __repr__(self):
        return "not("+self.args[0].__repr__()+")"

    def __str__(self):
        return "~"+self.args[0].__str__()


class And(LogicalOperator):
    """
    And Predicate
    """
    name = "and"
    arity = 2

    def __init__(self, arg1, arg2):
        super(And, self).__init__(arg1, arg2)
        self.args.append(arg1)
        self.args.append(arg2)

    def __repr__(self):
        return "and(" + self.args[0].__repr__() + ", " + \
               self.args[1].__repr__() + ")"

    def __str__(self):
        return "(" + self.args[0].__str__() + " & " + \
               self.args[1].__str__() + ")"


class Or(LogicalOperator):
    """
    Or Predicate
    """
    name = "or"
    arity = 2

    def __init__(self, arg1, arg2):
        super(Or, self).__init__(arg1, arg2)
        self.args.append(arg1)
        self.args.append(arg2)

    def __repr__(self):
        return "or(" + self.args[0].__repr__() + ", " \
               + self.args[1].__repr__() + ")"

    def __str__(self):
        return "(" + self.args[0].__str__() + " | " \
               + self.args[1].__str__() + ")"


class If(LogicalOperator):
    """
    Implies (Material Conditional) Predicate
    """
    name = "if"
    arity = 2

    def __init__(self, arg1, arg2):
        super(If, self).__init__(arg1, arg2)
        self.args.append(arg1)
        self.args.append(arg2)

    def __repr__(self):
        return "if(" + self.args[0].__repr__() + ", " \
               + self.args[1].__repr__() + ")"

    def __str__(self):
        return "(" + self.args[0].__str__() + " -> " \
               + self.args[1].__str__() + ")"


class Iff(LogicalOperator):
    """
    Equiv(alence) Predicate
    """
    name = "iff"
    arity = 2

    def __init__(self, arg1, arg2):
        super(Iff, self).__init__(arg1, arg2)
        self.args.append(arg1)
        self.args.append(arg2)

    def __repr__(self):
        return "iff(" + self.args[0].__repr__() + ", " \
               + self.args[1].__repr__() + ")"

    def __str__(self):
        return "(" + self.args[0].__str__() + " <-> " \
               + self.args[1].__str__() + ")"



"""
Automated Theorem Prover within Forseti
"""
# pylint: disable=fixme
from __future__ import print_function
from copy import deepcopy
from forseti.formula import Not, And, Or
from forseti import converter, parser
import forseti.util as util


class Prover(object):
    """
    Prover class
    """
    def __init__(self):
        self._cnf_list = []
        self.formulas = []
        self.goals = []
        self._goals = []

    def add_formula(self, statement):
        """

        :param statement:
        :return:
        """
        self._add_statement(deepcopy(statement), self.formulas)

    def add_goal(self, statement):
        """

        :param statement:
        :return:
        """
        self._add_statement(deepcopy(statement), self.goals)
        self._add_statement(Not(deepcopy(statement)), self._goals)

    @staticmethod
    def _add_statement(statement, add_to_list):
        """

        :param statement:
        :param add_to_list:
        :return:
        """
        if statement is None:
            raise TypeError("Statement cannot be None")
        statement = parser.parse(statement)
        statement = converter.convert_to_cnf(statement)
        add_to_list.append(statement)

    def run_prover(self):
        """

        :return:
        """
        if len(self.goals) == 0:
            # TODO: give this a better exception class
            raise Exception("You need at least one goal!")

        for formula in self.formulas:
            converts = self._convert_formula(formula)
            for convert in converts:
                if isinstance(convert, list):
                    self._cnf_list += [convert]
                else:
                    self._cnf_list.append([convert])
        for goal in self._goals:
            converts = self._convert_formula(goal)
            for convert in converts:
                if isinstance(convert, list):
                    self._cnf_list += [convert]
                else:
                    self._cnf_list.append([convert])

        self._tautology_elimination()
        for i in range(len(self._cnf_list)):
            self._cnf_list[i] = sorted(self._cnf_list[i])
        return self._resolve()

    @staticmethod
    def _convert_formula(formula):
        """
        Converts a CNF formula into lists for resolution.
        Ex:
        and(a,b) -> [[a], [b]]
        or(a,b) -> [[a,b]]
        and(or(a,b),c) -> [[a,b],[c]]

        :type formula: LogicalOperator or Symbol
        """
        cnf_list = [[formula]]
        all_checked = False
        break_from = False
        while not all_checked:
            all_checked = True
            for i in range(len(cnf_list)):
                for j in range(len(cnf_list[i])):
                    statement = cnf_list[i][j]
                    if isinstance(statement, And) or isinstance(statement, Or):
                        if isinstance(statement, And):
                            cnf_list.insert(i + 1, [statement.args[0]])
                            cnf_list.insert(i + 2, [statement.args[1]])
                        elif isinstance(statement, Or):
                            cnf_list[i].insert(j + 1, statement.args[0])
                            cnf_list[i].insert(j + 2, statement.args[1])

                        cnf_list[i].pop(j)
                        if len(cnf_list[i]) == 0:
                            cnf_list.pop(i)
                        break_from = True
                        all_checked = False
                        break
                if break_from is True:
                    break_from = False
                    break
        return cnf_list

    def _resolve(self):
        """

        :return:
        """
        i = 0
        checked = list()
        while i < len(self._cnf_list):
            j = i + 1
            while j < len(self._cnf_list):
                if [i, j] in checked:
                    j += 1
                    continue
                checked.append([i, j])
                have_resolve = False
                for k in range(len(self._cnf_list[i])):
                    atomic = self._cnf_list[i][k]
                    negation = util.negate_formula(atomic)
                    try:
                        ind = self._cnf_list[j].index(negation)
                        new_cnf = self._cnf_list[i][:k]
                        new_cnf += self._cnf_list[i][(k+1):]
                        for cnf in self._cnf_list[j][:ind]:
                            if cnf not in new_cnf:
                                new_cnf.append(cnf)
                        for cnf in self._cnf_list[j][(ind+1):]:
                            if cnf not in new_cnf:
                                new_cnf.append(cnf)
                        new_cnf.sort()
                        if len(new_cnf) == 0:
                            return True
                        if not util.is_tautology(new_cnf) and \
                           new_cnf not in self._cnf_list:
                            have_resolve = True
                            self._cnf_list.append(new_cnf)
                            checked.append([i, len(self._cnf_list)-1])
                            checked.append([j, len(self._cnf_list)-1])
                    except ValueError:
                        pass
                if have_resolve is True:
                    i = -1
                    break
                j += 1
            i += 1
        return False

    def _tautology_elimination(self):
        """

        :return:
        """
        i = 0
        while i < len(self._cnf_list):
            cnf = self._cnf_list[i]
            if util.is_tautology(cnf):
                self._cnf_list.pop(i)
                i -= 1
            i += 1



"""
Util functions utilized by Forseti library (or for testing it)
"""

from __future__ import print_function
from forseti.formula import Not, LogicalOperator, Symbol
import sys


def print_cnf_list(formulas, depth=0, end="\n", out=sys.stdout):
    """
    prints out the CNF list

    :param formulas:
    :return:
    """
    output = "["
    for i in formulas:
        if isinstance(i, list):
            output += print_cnf_list(i, depth=depth+1)
        else:
            output += str(i)
        output += ", "
    output = output[:-2] + "]"
    if depth > 0:
        return output
    else:
        out.write(output+end)


def negate_formula(formula):
    """
    Negate the formula (or removing the Not if it's the outermost Operator)

    :param formula: formula
    :return: Negated formula
    """
    if isinstance(formula, Not):
        return formula.args[0]
    elif isinstance(formula, Symbol) or isinstance(formula, LogicalOperator):
        return Not(formula)
    else:
        raise TypeError(str(formula) + " is not a valid formula "
                                       "(Symbol or Operator)")


def is_tautology(cnf):
    """
    Checks if a given CNF is a tautology
    :return boolean: True if it is a tautology, false otherwise
    """
    if not isinstance(cnf, list) or len(cnf) <= 1:
        return False
    i = 0
    while i < len(cnf):
        if negate_formula(cnf[i]) in cnf:
            return True
        i += 1
    return False



#!/usr/bin/env python
"""
setup routine
"""
from setuptools import setup, find_packages
import os


def read(*paths):
    """Build a file path from *paths* and return the contents."""
    with open(os.path.join(*paths), 'r') as open_file:
        return open_file.read()

setup(
    # Metadata
    name='forseti',
    version='0.6.0',
    description='Formal Logic Framework',
    long_description=read('README.md'),
    url='https://github.com/MasterOdin/forseti',
    download_url='https://pypi.python.org/pypi/forseti',
    license='MIT',
    author='Matthew Peveler',
    author_email='matt.peveler@gmail.com',

    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: MIT License',
        'Natural Language :: English',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
        'Operating System :: OS Independent',
        'Topic :: Scientific/Engineering :: Mathematics',
    ],

    # Dependencies
    tests_require=['nose', 'six'],

    # Contents
    packages=find_packages(exclude=['tests*']),

)



"""
Converter functions for Forseti statements (to CNF)
"""

from forseti.formula import LogicalOperator, Symbol, Not, And, Or, If, Iff


def convert_to_cnf(statement):
    """
    Convert a given statement to Conjective Normal Form (CNF)

    :param statement:
    :return: a CNF statement
    """
    if isinstance(statement, Symbol):
        return statement
    elif not isinstance(statement, LogicalOperator):
        raise TypeError(str(statement) + " is not a Symbol or LogicalOperator "
                                         "Object. Use forseti.parser first")

    statement = _convert_equiv(statement)
    statement = _convert_implies(statement)
    statement = _distribute_not(statement)

    while not _is_cnf(statement):
        statement = _distribute_or(statement)

    return statement


def _convert_equiv(statement):
    """

    :param statement:
    :return:
    """
    if isinstance(statement, Symbol):
        return statement

    args = statement.args
    for i in range(len(args)):
        args[i] = _convert_equiv(args[i])

    if isinstance(statement, Iff):
        statement = And(If(args[0], args[1]), If(args[1], args[0]))

    return statement


def _convert_implies(statement):
    """

    :param statement:
    :return:
    """
    if isinstance(statement, Symbol):
        return statement

    args = statement.args
    for i in range(len(args)):
        args[i] = _convert_implies(args[i])

    if isinstance(statement, If):
        statement = Or(Not(args[0]), args[1])

    return statement


def _distribute_not(statement):
    """

    :param statement:
    :return:
    """
    if isinstance(statement, Symbol):
        return statement

    args = statement.args
    if isinstance(statement, Not):
        if isinstance(args[0], And) or isinstance(args[0], Or):
            new_type = Or
            if isinstance(args[0], Or):
                new_type = And
            args[0] = new_type(Not(args[0].args[0]), Not(args[0].args[1]))
            statement = _distribute_not(args[0])
        elif isinstance(args[0], Not):
            statement = args[0].args[0]
            statement = _distribute_not(statement)
    else:
        for i in range(len(args)):
            args[i] = _distribute_not(args[i])
    return statement


def _distribute_or(statement):
    """

    :param statement:
    :return:
    """
    if isinstance(statement, Symbol):
        return statement

    args = statement.args
    if isinstance(statement, Or):
        for i in range(len(args)):
            if isinstance(args[i], And):
                j = (i + 1) % 2
                statement = And(Or(args[i].args[0], args[j]),
                                Or(args[i].args[1], args[j]))
                break

    args = statement.args
    for i in range(len(args)):
        args[i] = _distribute_or(args[i])

    return statement


def _is_cnf(statement, has_or=False):
    """

    :param statement:
    :param has_or:
    :return:
    """
    if isinstance(statement, Symbol):
        return True
    elif isinstance(statement, If) or isinstance(statement, Iff):
        return False
    elif isinstance(statement, Or):
        has_or = True
    elif isinstance(statement, And) and has_or is True:
        return False

    args = statement.args
    for i in range(len(args)):
        if _is_cnf(args[i], has_or) is False:
            return False

    return True



"""
Parser for converting plain-text to Forseti formula
"""

from forseti.formula import Symbol, Not, And, Or, If, Iff, LogicalOperator


def parse(statement):
    """
    Parse a plain-text string into Forseti formula

    TODO: raise exception on invalid formulas

    :param statement:
    :return: Predicate
    """
    if isinstance(statement, LogicalOperator) or isinstance(statement, Symbol):
        return statement

    original = statement
    statement = statement.replace(" ", "")

    if statement.count("(") > statement.count(")"):
        raise SyntaxError("Invalid formula (check parentheses): " + original)

    if len(statement) == 0 or statement == "()":
        raise SyntaxError("Invalid formula: " + original)

    while statement[0] == "(" and statement[-1] == ")":
        statement = statement[1:-1]

    if statement.lower().startswith("and"):
        parse_type = And
    elif statement.lower().startswith("or"):
        parse_type = Or
    elif statement.lower().startswith("iff"):
        parse_type = Iff
    elif statement.lower().startswith("if"):
        parse_type = If
    elif statement.lower().startswith("not"):
        parse_type = Not
    else:
        parse_type = Symbol
    parsed_statement = _parse_statement(statement, parse_type)
    if parsed_statement is None:
        raise SyntaxError("Invalid formula: " + original)
    else:
        return parsed_statement


def _parse_statement(statement, parse_type):
    """
    Break the string into arguments to pass into the given statement

    :param statement:
    :param statement:
    :return:
    """
    if parse_type == Symbol:
        symbol_string = ""
        for char in statement:
            if not char.isalnum():
                return None
            symbol_string += char
        return Symbol(symbol_string)
    else:
        statement = statement[len(parse_type.name):]
        if len(statement) > 2 and statement[0] == "(" and statement[-1] == ")":
            statement = statement[1:-1]
            arg_list = []
            status = _get_arg_list(arg_list, statement)
            if status is False or len(arg_list) != parse_type.arity:
                return None
            return parse_type(*arg_list)
        else:
            return None


def _get_arg_list(arg_list, string):
    """
    get the argument list of a 2+ argument statement

    :param arg_list:
    :param string:
    :return:
    """
    i = 0
    open_p = 0
    arg = ""
    while i < len(string):
        char = string[i]
        if char == "(":
            open_p += 1

        elif char == ")" and open_p >= 0:
            open_p -= 1

        if not char.isalnum() and ((char != "," and char != ")") and open_p == 0):
            return False

        if char == "," and open_p == 0:
            arg_list.append(parse(arg))
            arg = ""
        else:
            arg += char
        i += 1
    arg_list.append(parse(arg))
    return True
